/*
 * milter-spamc.c
 *
 * Copyright 2003, 2004 by Anthony Howe. All rights reserved.
 *
 * The following should be added to the sendmail.mc file:
 *
 *	INPUT_MAIL_FILTER(
 *		`milter-spamc',
 *		`S=unix:/var/lib/milter-spamc/socket, T=S:30s;R:3m'
 *	)dnl
 */

/***********************************************************************
 *** Leave this header alone. Its generate from the configure script.
 ***********************************************************************/

#include "config.h"

/***********************************************************************
 *** You can change the stuff below if the configure script doesn't work.
 ***********************************************************************/

#ifndef RUN_AS_USER
#define RUN_AS_USER			"smmsp"
#endif

#ifndef RUN_AS_GROUP
#define RUN_AS_GROUP			"smmsp"
#endif

#ifndef STATEDIR
#define STATEDIR			"/var/lib/" MILTER_NAME
#endif

#ifndef DEFAULT_SPAMD_HOST
#define DEFAULT_SPAMD_HOST		"127.0.0.1"
#endif

#ifndef DEFAULT_SPAMD_PORT
#define DEFAULT_SPAMD_PORT		783
#endif

#ifndef DEFAULT_SOCKET_TIMEOUT
#define DEFAULT_SOCKET_TIMEOUT		120
#endif

#ifndef DEFAULT_HEADER_BUFFER_SIZE
#define DEFAULT_HEADER_BUFFER_SIZE	8192
#endif

#ifndef DEFAULT_SPAM_SUBJECT_TAG
#define DEFAULT_SPAM_SUBJECT_TAG	"[SPAM]"
#endif

#ifndef DEFAULT_SPAM_LEVEL_CHARACTER
#define DEFAULT_SPAM_LEVEL_CHARACTER	'x'
#endif

/*
 * Choose between CHECK, REPORT_IFSPAM, and REPORT.  
 * PROCESS and SYMBOLS not supported.
 */
#ifndef DEFAULT_SPAMC_COMMAND
#define DEFAULT_SPAMC_COMMAND		"REPORT"
#endif

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

/* Re-assert this macro just in case. May cause a compiler warning. */
#define _REENTRANT	1

#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/stat.h>
#include <netdb.h>

#include <com/snert/lib/version.h>
#include <com/snert/lib/io/Socket.h>
#include <com/snert/lib/mail/smf.h>
#include <com/snert/lib/mail/smdb.h>
#include <com/snert/lib/util/Buf.h>
#include <com/snert/lib/util/Text.h>
#include <com/snert/lib/util/getopt.h>

#if LIBSNERT_MAJOR < 1 || LIBSNERT_MINOR < 39
# error "LibSnert/1.39 or better is required"
#endif

#if defined(LIBSNERT_WITH_BERKELEY_DB) && defined(MILTER_WITH_BERKELEY_DB)
# define HAVE_BERKELEY_DB
#else
# undef HAVE_BERKELEY_DB
#endif

/***********************************************************************
 *** Constants
 ***********************************************************************/

#define	TAG_FORMAT		"%05d %s: "
#define	TAG_ARGS		data->work.cid, data->work.qid

#define COPY_TO			0
#define REDIRECT_TO		1
#define REDIRECT_TO_FORMAT	2

/***********************************************************************
 *** Global Variables
 ***********************************************************************/

typedef struct {
	int isLocal;				/* rcpt_mailer == 'local or 'cyrusv2' */
	long length;				/* address length */
	long capacity;				/* address buffer capacity */
	char *address;				/* recipient address with < and > */
} *Rcpt;

typedef struct {
	smfWork work;
	Socket server;				/* per message */
	Vector rcpts;				/* per message */
	Buf headers;				/* per message */
	time_t now;				/* per message */
	size_t bytesSent;			/* per message */
	int localRcptCount;			/* per message */
	int untrustedRcptCount;			/* per message */
	int hasSubject;				/* per message */
	int hasSpamFlag;			/* per message */
	int hasSpamLevel;			/* per message */
	int hasSpamReport;			/* per message */
	int hasSpamStatus;			/* per message */
	int precedence;				/* per message */
	long indexReceived;			/* per message */
	char *localRcpt;			/* per message */
	char helo[DOMAIN_LENGTH+1];		/* per connection */
	char line[SMTP_LINE_SIZE+1];		/* general purpose */
	char subject[SMTP_LINE_SIZE+1];		/* subject header */
	char client_name[DOMAIN_LENGTH+1];	/* per connection */
	char client_addr[5+IPV6_LENGTH+1];	/* per connection */
} *workspace;

static smfInfo milter = {
	MILTER_NAME,
	MILTER_VERSION,
	MILTER_COPYRIGHT,
	RUN_AS_USER,	
	RUN_AS_GROUP,
	STATEDIR,
	NULL,
	SMF_STDIO_CLOSE
};

static long socketTimeout;
static char *spamcCommand;
static int alwaysAddReport;
static int forceSpamdUser;
static int redirectSpam;
static int spamdPort = DEFAULT_SPAMD_PORT;
static int untrustedRcptLimit;
static int checkPrecedenceList;
static int appendScoreToSubjectTag;
static char *spamdHost = DEFAULT_SPAMD_HOST;
static char *spamdUser;
static char *copyAllTo;
static char *copyHamTo;
static char *copySpamTo;
static char *rcptReformat;
static char *subjectTag;
static char *subjectScoreTag;
static char *sendmailCf;
static long maxBytes = 64;
static long rejectExcessScore;
static long discardExcessScore;

#ifdef SMFIS_REJECT_AND_DELIVER
static long rejectAndDeliverExcessScore;
#endif

static char getoptString[] = "ACpSa:b:B:f:F:c:d:D:h:K:l:r:R:s:t:T:u:U:v:H:P:?";

static char usageMessage[] =
"\n"
"milter-spamc [-A|-C][-p][-a email][-b email|-B email][-f cf][-F format]\n"
"         [-c email][-d excess][-K kilo][-l limit][-r excess][-R excess]\n"
"         [-s tag][-S][-t timeout][-T timeout][-u user|-U user][-v detail]\n"
"         [-H host][-P port] socket\n"
"\n"
"-a email\tcopy all mail to email address\n"
"-A\t\talways add the X-Spam-Report: header report of rules triggered\n"
"-b email\tcopy junk mail to email address\n"
"-B email\tredirect junk mail to email address\n"
"-c email\tcopy clean mail to email address\n"
"-C\t\tcheck for spam only, disable X-Spam-Report: header completely\n"
"-d excess\tdiscard spam messages that exceed the threshold by N\n"
"-f cf\t\tpathname of sendmail.cf file for access database whitelists\n"
"-F format\tredirect junk mail to reformated recipient addresses\n"
"-h\t\tthis usage summary\n"
"-H host\t\tspamd host name or IP\n"
"-K kilo\t\thow many KB to pass to SpamAssassin, 64 default, 0 unlimited\n"
"-l limit\tlimit number of untrusted recipients per message\n"
"-p\t\tdiscard instead of reject if Precedence is list or lower\n"
"-P port\t\tspamd port address, -1 to disable spamd filtering\n"
"-r excess\treject spam that exceeds the threshold by N, overrides delivery\n"
"-R excess\treject spam that exceeds the threshold by N, allow delivery\n"
"-s tag\t\tspam subject tag\n"
"-S\t\tappend the spam score to the spam subject tag\n"
"-t timeout\tsendmail/milter I/O timeout in seconds, 0 indefinite\n"
"-T timeout\tmilter/spamd I/O timeout in seconds, 0 indefinite\n"
"-u user\t\tdefault user account to process messages with\n"
"-U user\t\tforce message processing with this user account\n"
"-v detail\tcomma separated list of what to write to the mail log:\n"
"\t\t0, all, info, trace, parse, debug, dialog, state, dns\n"
"\t\tcache, database, socket-fd, socket-all, libmilter\n"
"\n"
MILTER_VERSION " " MILTER_COPYRIGHT "\n"
"\n"
;

#define X_SCANNED_BY			"X-Scanned-By"
#define X_SPAM_FLAG			"X-Spam-Flag"
#define X_SPAM_LEVEL			"X-Spam-Level"
#define X_SPAM_REPORT			"X-Spam-Report"
#define X_SPAM_STATUS			"X-Spam-Status"
#define X_ORIGINAL_RECIPIENT		"X-Original-Recipient"

/***********************************************************************
 *** Support routines.
 ***********************************************************************/

static void atExitCleanUp(void);

/*
 * Write buffer, logging only errors.
 */
int
writebuffer(workspace data, char *buf, long size)
{
	int error;
	int rc = 0;
	long nbytes, length;

	/* Do not block on sending to the spamd server just yet. */
	(void) SocketSetNonBlocking(data->server, 1);

	for (length = 0; length < size; length += nbytes) {
		if ((nbytes = SocketWrite(data->server, buf + length, size - length)) < 0) {
			error = SocketGetError(data->server);
			syslog(LOG_ERR, TAG_FORMAT "SocketWrite() error: %s (%d)", TAG_ARGS, strerror(error), error);
			goto error0;
		}

		/* Now wait for the output to be sent to the spamd server. */
		if (SocketWaitForOutput(data->server, socketTimeout) != 0) {
			syslog(LOG_ERR, TAG_FORMAT "timeout before output sent to SPAMD server", TAG_ARGS);
			goto error0;
		}
	}

	rc = 1;
error0:
	(void) SocketSetNonBlocking(data->server, 0);

	return rc;
}

/*
 * Log line to be sent then write line.
 */
int
writeline(workspace data, char *line, long size)
{
	if (smfLogDetail & SMF_LOG_DIALOG)
		syslog(LOG_DEBUG, TAG_FORMAT "> %s", TAG_ARGS, line);

	return writebuffer(data, line, size < 0 ? strlen(line) : size);
}

/*
 * Log amount of data sent then write data.
 */
int
writechunk(workspace data, char *chunk, long size)
{
	if (smfLogDetail & SMF_LOG_DIALOG)
		syslog(LOG_DEBUG, TAG_FORMAT "> [%ld bytes of data]", TAG_ARGS, size);

	return writebuffer(data, chunk, size);
}

int
readline(workspace data, char *line, int linesize)
{
	long length;
	int rc, error;

	rc = -1;

	if (linesize < sizeof (long)) {
		syslog(LOG_CRIT, TAG_FORMAT "line buffer too small (%d)", TAG_ARGS, linesize);
		atExitCleanUp();
	}

	error = SocketWaitForInput(data->server, socketTimeout);

	if (error == ETIMEDOUT) {
		syslog(LOG_ERR, TAG_FORMAT "timeout before input from SPAMD server", TAG_ARGS);
		goto error0;
	}

	if (error != 0) {
		syslog(LOG_ERR, TAG_FORMAT "input error from SPAMD server: %s (%d)", TAG_ARGS, strerror(error), error);
		goto error0;
	}

	length = SocketReadLine(data->server, line, linesize);

	if ((error = SocketGetError(data->server)) != 0 || length == -1) {
		syslog(LOG_ERR, TAG_FORMAT "read error from SPAMD server: %s (%d)", TAG_ARGS, strerror(error), error);
		goto error0;
	}

	if (smfLogDetail & SMF_LOG_DIALOG)
		syslog(LOG_DEBUG, TAG_FORMAT "< %s", TAG_ARGS, line);

	rc = length;
error0:
	return rc;
}

/***********************************************************************
 *** Handlers
 ***********************************************************************/

/*
 * Open and allocate per-connection resources.
 */
static sfsistat
filterOpen(SMFICTX *ctx, char *client_name, _SOCK_ADDR *raw_client_addr)
{
	int rc;
	workspace data;
	unsigned short cid;
	char client_addr[5+IPV6_LENGTH+1];	
	
	cid = smfOpenProlog(ctx, client_name, raw_client_addr, client_addr, sizeof (client_addr));	

	if (smfLogDetail & SMF_LOG_TRACE)
		syslog(LOG_DEBUG, TAG_FORMAT "filterOpen(%lx, '%s', [%s])", cid, smfNoQueue, (long) ctx, client_name, client_addr);

	if ((data = calloc(1, sizeof *data)) == NULL)
		goto error0;

	data->work.ctx = ctx;
	data->work.cid = cid;
	data->work.qid = smfNoQueue;
	
	if ((data->headers = BufCreate(DEFAULT_HEADER_BUFFER_SIZE)) == (Buf) 0)
		goto error1;

	if ((data->rcpts = VectorCreate(10)) == NULL)
		goto error2;

	VectorSetDestroyEntry(data->rcpts, free);

	if (smfi_setpriv(ctx, (void *) data) == MI_FAILURE) {
		syslog(LOG_ERR, TAG_FORMAT "failed to save workspace", TAG_ARGS);
		goto error3;
	}

	snprintf(data->client_name, sizeof (data->client_name), "%s", client_name);
	snprintf(data->client_addr, sizeof (data->client_addr), "%s", client_addr);

	(void) smfAccessHost(&data->work, MILTER_NAME "-connect:", client_name, client_addr, SMDB_ACCESS_OK);

	return SMFIS_CONTINUE;
error3:
	VectorDestroy(data->rcpts);
error2:
	BufDestroy(data->headers);
error1:
	free(data);
error0:
	return smfNullWorkspaceError("filterOpen");
}

static sfsistat
filterHelo(SMFICTX * ctx, char *helohost)
{
	workspace data;

	if ((data = (workspace) smfi_getpriv(ctx)) == NULL)
		return smfNullWorkspaceError("filterHelo");

	if (smfLogDetail & SMF_LOG_TRACE)
		syslog(LOG_DEBUG, TAG_FORMAT "filterHelo(%lx, '%s')", TAG_ARGS, (long) ctx, helohost);

	TextCopy(data->helo, sizeof(data->helo), helohost);

	return SMFIS_CONTINUE;
}

static sfsistat
filterMail(SMFICTX *ctx, char **args)
{
	workspace data;

	if ((data = (workspace) smfi_getpriv(ctx)) == NULL)
		return smfNullWorkspaceError("filterMail");

	if ((data->work.qid = smfi_getsymval(ctx, "i")) == NULL)
		data->work.qid = smfNoQueue;

	if (smfLogDetail & SMF_LOG_TRACE)
		syslog(LOG_DEBUG, TAG_FORMAT "filterMail(%lx, %lx) MAIL='%s'", TAG_ARGS, (long) ctx, (long) args, args[0]);
		
	/* Reset per-message variables, since this function can be
	 * called several times per message or connection, ie.
	 * MAIL-RSET-MAIL or MAIL-RCPT-DATA-MAIL sequences.
	 */

	VectorRemoveSome(data->rcpts, 0, VectorLength(data->rcpts));
	BufSetLength(data->headers, 0);
	data->untrustedRcptCount = 0;
	data->localRcptCount = 0;

	if (data->localRcpt != spamdUser)
		free(data->localRcpt);

	data->work.skipMessage = spamdPort < 0 || data->work.skipConnection;
	data->precedence = PRECEDENCE_FIRST_CLASS;
	data->indexReceived = -1;
	data->localRcpt = NULL;
	data->bytesSent = 0;

	(void) time(&data->now);

	/* Make sure its closed before we start the next message. */
	SocketClose(data->server);
	data->server = NULL;

	if (smfAccessMail(&data->work, MILTER_NAME "-from:", args[0], SMDB_ACCESS_UNKNOWN) == SMDB_ACCESS_ERROR)
		return SMFIS_REJECT;

	return SMFIS_CONTINUE;
}

static sfsistat
filterRcpt(SMFICTX *ctx, char **args)
{
	workspace data;
	char *rcpt_addr, *rcpt_host, *rcpt_mailer;

	if ((data = (workspace) smfi_getpriv(ctx)) == NULL)
		return smfNullWorkspaceError("filterRcpt");

	rcpt_addr = smfi_getsymval(ctx, "{rcpt_addr}");
	rcpt_host = smfi_getsymval(ctx, "{rcpt_host}");
	rcpt_mailer = smfi_getsymval(ctx, "{rcpt_mailer}");
		
	if (smfLogDetail & (SMF_LOG_INFO|SMF_LOG_TRACE))
		syslog(
			LOG_DEBUG, TAG_FORMAT "RCPT='%s' rcpt_addr='%s' rcpt_host='%s' rcpt_mailer='%s'",
			TAG_ARGS, args[0], rcpt_addr, rcpt_host, rcpt_mailer
		);

	if (smfAccessRcpt(&data->work, MILTER_NAME "-to:", args[0]) == SMDB_ACCESS_ERROR)
		return SMFIS_REJECT;

	/* TEST if the untrustedRcptCount has exceed the -l limit. */
	if (0 < untrustedRcptLimit && untrustedRcptLimit <= data->untrustedRcptCount)
		return smfReply(&data->work, 452, "4.5.3", "too many untrusted recipients");
/*		return smfReply(&data->work, 421, "4.3.2", "close connection"); */

	data->work.rcpt->isLocal = *rcpt_host == '\0' || strncmp(rcpt_mailer, "cyrus", 5) == 0;

	/* Remember last local user. Ideally we should also attempt to
	 * expand mail aliases, but that means lots of Berkeley DB issues,
	 * so for the time being we'll ignore it and hope its not that
	 * important.
	 *
	 * Normally {rcpt_mailer} == 'local' or 'cyrusv2', {rcpt_host} == ''
	 * and {rcpt_addr} == account. However, some sites do some special
	 * rulesets for cyrus or cyrusv2 where the {rcpt_host} != '', yet
	 * the user is a local account.
	 */
	if (data->work.rcpt->isLocal) {
		int hasEntry;
		
		if (pthread_mutex_lock(&smfMutex))
			syslog(LOG_ERR, "mutex lock in filterRcpt() failed: %s (%d)", strerror(errno), errno);	

		hasEntry = getpwnam(rcpt_addr) != NULL;

		if (pthread_mutex_unlock(&smfMutex))
			syslog(LOG_ERR, "mutex unlock in filterRcpt() failed: %s (%d)", strerror(errno), errno);	

		if (hasEntry) {
			/* Free NULL or a previous local recipient. */
			free(data->localRcpt);

			if (smfLogDetail & SMF_LOG_TRACE)
				syslog(LOG_DEBUG, TAG_FORMAT "address %s has an account", TAG_ARGS, args[0]);

			/* Assign local recipient. Don't worry if NULL, handled in filterBody(). */
			data->localRcpt = TextDup(rcpt_addr);
		}
	}

	data->localRcptCount += data->work.rcpt->isLocal;
	data->untrustedRcptCount++;

	if (VectorAdd(data->rcpts, data->work.rcpt))
		return smfReply(&data->work, 452, "4.3.2", "out of memory, cannot add recipient to list");
		
	data->work.rcpt = NULL;

	return SMFIS_CONTINUE;
}

static sfsistat
filterHeader(SMFICTX *ctx, char *name, char *value)
{
	int i;
	workspace data;

	if ((data = (workspace) smfi_getpriv(ctx)) == NULL)
		return smfNullWorkspaceError("filterHeader");

	if (smfLogDetail & SMF_LOG_TRACE)
		syslog(LOG_DEBUG, TAG_FORMAT "filterHeader(%lx, '%s', '%s')", TAG_ARGS, (long) ctx, name, value);

	if (data->work.skipMessage || name == NULL || value == NULL)
		return SMFIS_CONTINUE;

	if (data->indexReceived < 0 && TextInsensitiveCompare(name, "Received") == 0) {
		data->indexReceived = BufLength(data->headers);
		if (smfLogDetail & SMF_LOG_DEBUG)
			syslog(LOG_DEBUG, TAG_FORMAT "indexReceived=%ld", TAG_ARGS, data->indexReceived);
	}

	if (TextInsensitiveCompare(name, "Subject") == 0) {
		(void) strncpy(data->subject, value, sizeof (data->subject) - strlen(subjectTag) - 1);
		data->subject[sizeof (data->subject) - strlen(subjectTag) - 1] = '\0';
		data->hasSubject = 1;
	}

	/* There appears to be a bug in libmilter. smfi_chgheader() is
	 * suppose to change or add the header if missing, but it don't
	 * work as advertised, so we have to track the precense of these
	 * headers and call smfi_chgheader() or smfi_addheader() as required.
	 */
	if (TextInsensitiveCompare(name, X_SPAM_FLAG) == 0) {
		data->hasSpamFlag = 1;
		return SMFIS_CONTINUE;
	}		
	if (TextInsensitiveCompare(name, X_SPAM_LEVEL) == 0) {
		data->hasSpamLevel = 1;
		return SMFIS_CONTINUE;
	}
	if (TextInsensitiveCompare(name, X_SPAM_REPORT) == 0) {
		data->hasSpamReport = 1;
		return SMFIS_CONTINUE;
	}
	if (TextInsensitiveCompare(name, X_SPAM_STATUS) == 0) {
		data->hasSpamStatus = 1;
		return SMFIS_CONTINUE;
	}
	
	if (TextInsensitiveCompare(name, "Precedence") == 0) {
		for (i = 0; i <= PRECEDENCE_SPECIAL_DELIVERY; i++) {
			if (TextInsensitiveCompare(value, smfPrecedence[i]) == 0) {
				data->precedence = i;
				break;
			}
		}
		return SMFIS_CONTINUE;
	}

	(void) BufAddString(data->headers, name);
	(void) BufAddBytes(data->headers, ": ", 0, 2);
	(void) BufAddString(data->headers, value);
	(void) BufAddBytes(data->headers, "\r\n", 0, 2);

	return SMFIS_CONTINUE;
}

static sfsistat
filterEndHeaders(SMFICTX *ctx)
{
	long length, n;
	workspace data;

	if ((data = (workspace) smfi_getpriv(ctx)) == NULL)
		return smfNullWorkspaceError("filterEndHeaders");

	if (smfLogDetail & SMF_LOG_TRACE)
		syslog(LOG_DEBUG, TAG_FORMAT "filterEndHeaders(%lx)", TAG_ARGS, (long) ctx);

	if (data->work.skipMessage)
		return SMFIS_CONTINUE;

	(void) BufAddBytes(data->headers, "\r\n", 0, 2);

	/* Insert a simulated Received: header for this server into the
	 * header block being sent to spamd. It appears that this header
	 * is never given to the milter and without it, some tests in
	 * SpamAssassin concerning Received headers are skewed.
	 *
	 * This simulated header will be a stripped down version of
	 * what Sendmail would insert.
	 */
	length = snprintf(
		data->line, SMTP_LINE_SIZE, "Received: from %s (%s [%s]) by %s id %s; ", 
		data->helo, data->client_name, data->client_addr, smfi_getsymval(ctx, "{if_name}"), data->work.qid
	);			

	if (0 < (n = TimeStampAdd(data->line + length, SMTP_LINE_SIZE - length - 2))) {
		data->line[length + n    ] = '\r';
		data->line[length + n + 1] = '\n';

		if (data->indexReceived < 0)
			data->indexReceived = 0;

		(void) BufInsertBytes(data->headers, data->indexReceived, data->line, 0, length + n + 2);

		if (smfLogDetail & SMF_LOG_DEBUG)
			syslog(LOG_DEBUG, TAG_FORMAT "after length=%ld headers {%s}", TAG_ARGS, BufLength(data->headers), BufBytes(data->headers));		
	}

	return SMFIS_CONTINUE;
}

static int
spamdConnect(workspace data, size_t size)
{
	long length;
	
	/* Start sending the headers for processing to SpamAssassin now. 
	 * If there is no body after the headers filterBody() will not
	 * be called.
	 */
	data->server = SocketOpenTcpClientWait(spamdHost, spamdPort, socketTimeout);

	if (data->server == (Socket) 0) {
		syslog(LOG_ERR, TAG_FORMAT "failed to connect to spamd %s,%d", TAG_ARGS, spamdHost, spamdPort);
		goto error0;
	}

	if (smfLogDetail & SMF_LOG_DIALOG)
		syslog(LOG_DEBUG, TAG_FORMAT "SPAMD Content-Length=%ld", TAG_ARGS, BufLength(data->headers) + size);

	/* Ask for a CHECK or REPORT only and we'll handle adding the X-Spam-* headers. */
	length = snprintf(
		data->line, SMTP_LINE_SIZE, "%s SPAMC/1.2\r\nContent-Length: %ld\r\n",
		spamcCommand, BufLength(data->headers) + size
	);
	if (!writeline(data, data->line, length))
		goto error1;

	/* If the spamd user is defined, then we'll use it if forced or
	 * multiple recipients given. Otherwise fall back on the last
	 * defined local recipient if any.
	 */
	if (spamdUser != (char *) 0 && (forceSpamdUser || 1 < VectorLength(data->rcpts))) {
		/* Free NULL or a previous local user. */
		free(data->localRcpt);
		data->localRcpt = (char *) 0;
	}

	/* When no local recipient defined, then fallback on the default
	 * spamd user, which might be undefined.
	 */
	if (data->localRcpt == (char *) 0)
		data->localRcpt = spamdUser;

	if (data->localRcpt != (char *) 0) {
		if (smfLogDetail & SMF_LOG_DIALOG)
			syslog(LOG_DEBUG, TAG_FORMAT "SPAMD User=%s", TAG_ARGS, data->localRcpt);
		length = snprintf(data->line, SMTP_LINE_SIZE, "User: %s\r\n", data->localRcpt);
		if (!writeline(data, data->line, length))
			goto error1;
	}

	if (!writeline(data, "\r\n", 2))
		goto error1;

	if (!writechunk(data, BufBytes(data->headers), BufLength(data->headers)))
		goto error1;

	return 0;
error1:
	SocketClose(data->server);
	data->server = (Socket) 0;
error0:
	return -1;
}

static sfsistat
filterBody(SMFICTX *ctx, unsigned char *chunk, size_t size)
{
	workspace data;

	if ((data = (workspace) smfi_getpriv(ctx)) == NULL)
		return smfNullWorkspaceError("filterBody");

	if (smfLogDetail & SMF_LOG_TRACE)
		syslog(LOG_DEBUG, TAG_FORMAT "filterBody(%lx, '%.20s...', %lu) maxBytes=%ld bytesSent=%lu", TAG_ARGS, (long) ctx, chunk, size, maxBytes, data->bytesSent);

	if (data->work.skipMessage)
		goto error0;

	if (0 < maxBytes && maxBytes <= data->bytesSent)
		goto error0;

	/* On first chunk, open connection and send headers. */
 	if (data->bytesSent == 0 && spamdConnect(data, size))
 		goto error0;

 	/* Keep track of how much of the body we process. */
 	data->bytesSent += size;
 	
	if (!writechunk(data, chunk, size))
		goto error1;

	if (0 < maxBytes && maxBytes <= data->bytesSent) {
		/* Signal EOF to SPAMD so that it can begin processing now. 
		 * This appears to dramatically improve spamd performance.
		 */
		if (smfLogDetail & SMF_LOG_DIALOG)
			syslog(LOG_DEBUG, TAG_FORMAT "EOF (1) -> spamd", TAG_ARGS);			
		SocketShutdownNow(data->server, SHUT_WR);
	}

	return SMFIS_CONTINUE;
error1:
	SocketClose(data->server);
	data->server = (Socket) 0;
error0:
	return SMFIS_CONTINUE;
}

static sfsistat
filterEndMessage(SMFICTX *ctx)
{
	workspace data;
	ParsePath *rcpt;
	float score, threshold;
	int i, isSpam, length, iscore;
	const char *if_name, *if_addr;

	if ((data = (workspace) smfi_getpriv(ctx)) == NULL)
		return smfNullWorkspaceError("filterEndMessage");

	if (smfLogDetail & SMF_LOG_TRACE)
		syslog(LOG_DEBUG, TAG_FORMAT "filterEndMessage(%lx)", TAG_ARGS, (long) ctx);

	/* Send a copy of all mail to the archive address. */
	if (copyAllTo != (char *) 0 && smfi_addrcpt(ctx, copyAllTo) == MI_FAILURE)
		syslog(LOG_ERR, TAG_FORMAT "copy-all-to <%s> failed", TAG_ARGS, copyAllTo);

	/* Skip the remainder of this function. */
	if (data->work.skipMessage)
		return SMFIS_CONTINUE;

	if (data->bytesSent == 0) {
		/* No spamd session started yet, because of no
		 * body content; send only the headers now. 
		 */
		if (spamdConnect(data, 0))
			goto error0;
	} else if (data->server == NULL) {
		/* Previous connection error in filterBody(). */
		goto error0;
	}

	/* Signal EOF to spamd so that it can start processing. 
	 * Note this might have been sent already in filterBody().
	 */
	if (smfLogDetail & SMF_LOG_DIALOG)
		syslog(LOG_DEBUG, TAG_FORMAT "EOF (2) -> spamd", TAG_ARGS);			
	SocketShutdownNow(data->server, SHUT_WR);

	SocketSetNonBlocking(data->server, 1); 

	/* Fetch spamd result and add it to the headers. */
	if (readline(data, data->line, SMTP_LINE_SIZE) <= 0) {
		syslog(LOG_ERR, TAG_FORMAT "SPAMD status line failure", TAG_ARGS);
		goto error0;
	}

	if (strncmp(data->line, "SPAMD/1.1 0 EX_OK", sizeof ("SPAMD/1.1 0 EX_OK")) != 0) {
		syslog(LOG_ERR, TAG_FORMAT "SPAMD failure: %s", TAG_ARGS, data->line);
		goto error0;
	}

	/* Set to -2 for undefined, since TextInsensitiveStartsWith() 
	 * returns -1 on no match or a zero/postive length for a match.
	 */
	isSpam = -2;
	
	while (0 < readline(data, data->line, SMTP_LINE_SIZE)) {
		if (0 < TextInsensitiveStartsWith(data->line, "Spam:")) {
			/* And the verdict is? */
			isSpam = 0 < TextInsensitiveStartsWith(data->line, "Spam: True");
			(void) smfHeaderSet(ctx, X_SPAM_FLAG, isSpam ? "YES" : "NO", 1, data->hasSpamFlag);

			/* Extract the score. */
			if (sscanf(data->line, "%*[^;]; %f / %f", &score, &threshold) != 2) {
				syslog(LOG_ERR, TAG_FORMAT "Spam: header parse error: %s", TAG_ARGS, data->line);
				goto error0;
			}
		}
		
		/* Ignore any other headers. */
	}

	if ((if_name = smfi_getsymval(ctx, "{if_name}")) == NULL)
		if_name = smfUndefined;
	if ((if_addr = smfi_getsymval(ctx, "{if_addr}")) == NULL)
		if_addr = "0.0.0.0";

	length = snprintf(data->line, sizeof (data->line), MILTER_VERSION " (%s [%s]); ", if_name, if_addr);			
	length += TimeStampAdd(data->line + length, SMTP_LINE_SIZE - length);
	(void) smfi_addheader(ctx, X_SCANNED_BY, data->line);
	
	if (isSpam == -2) {
		syslog(LOG_ERR, TAG_FORMAT "missing Spam: result header", TAG_ARGS);
		goto error0;
	}

	/* Add a pretty header. I think this one is useless. */
	(void) sprintf(data->line, "%s, hits=%.2f required=%.2f", isSpam ? "YES" : "NO", score, threshold);
	(void) smfHeaderSet(ctx, X_SPAM_STATUS, data->line, 1, data->hasSpamStatus);

	/* Add another header that can be parsed with by a regular expression.
	 * This one is useful for filtering different levels of spam. Note that
	 * we initially use a non-regular-expression character for writing the
	 * level.
	 */
	if (SMTP_LINE_SIZE <= score)
		/* X-Spam-Level: xxx...x\r\n\0 */
		score = SMTP_LINE_SIZE - sizeof (X_SPAM_LEVEL) - 5;

	iscore = (int) score;
	for (i = 0; i < iscore; i++)
		data->line[i] = DEFAULT_SPAM_LEVEL_CHARACTER;
	data->line[i] = '\0';
	
	(void) smfHeaderSet(ctx, X_SPAM_LEVEL, data->line, 1, data->hasSpamLevel);

	/* REUSE this buffer to collect the report. */
	BufSetLength(data->headers, 0);

	/* SpamAssassin 2.5x prefixed a redundant score line for a report. */
	length = readline(data, data->line, SMTP_LINE_SIZE);
	if (strspn(data->line, "1234567890-./") != length) {
		/* Remove trailing whitespace. */
		while (0 < length-- && isspace(data->line[length]))
			data->line[length] = '\0';

		/* Only add non-empty lines. */
		if (data->line[0] != '\0') {
			BufAddString(data->headers, data->line);
			BufAddString(data->headers, "\n  ");
		}
	}

	/* Collect report lines until read error or EOF. */
	while (0 <= (length = readline(data, data->line, SMTP_LINE_SIZE))) {
		/* Remove trailing whitespace. */
		while (0 < length-- && isspace(data->line[length]))
			data->line[length] = '\0';

		if (length <= 0) {
			/* Replace a empty line with something that looks blank. */
			BufAddString(data->headers, "____\n  ");
		} else {
			BufAddString(data->headers, data->line);
			BufAddString(data->headers, "\n  ");
		}
	}

	/* Done with spamd session. */
	SocketClose(data->server);
	data->server = NULL;

	/* Remove our trailing whitespace "\r\n  ". */
	BufSetLength(data->headers, BufLength(data->headers)-3);

	if (alwaysAddReport)
		(void) smfHeaderSet(ctx, X_SPAM_REPORT, BufBytes(data->headers), 1, data->hasSpamReport);

	if (isSpam) {
		if (!alwaysAddReport)
			(void) smfHeaderSet(ctx, X_SPAM_REPORT, BufBytes(data->headers), 1, data->hasSpamReport);

		/* Tag the subject if not already. */
		if (TextInsensitiveStartsWith(data->subject, subjectTag) < 0) {
			if (appendScoreToSubjectTag)			
				(void) snprintf(data->line, sizeof data->line, "%s (%.2f/%.2f) %s", subjectTag, score, threshold, data->subject);
			else
				(void) snprintf(data->line, sizeof data->line, "%s %s", subjectTag, data->subject);

			(void) smfHeaderSet(ctx, "Subject", data->line, 1, data->hasSubject);
		}

		if (copySpamTo != (char *) 0) {
			/* The -F option can only redirect to individual mail folders when
			 * the entire recipient list is local. Otherwise send to the system
			 * spam box for processing.
			 *
			 * The reason for this restriction is that +detail information is 
			 * site specific and so adding this information into outbound mail
			 * is useless. Also when the mail is a mix of local and outbound mail
			 * (for example local user sending to multiple addresses locally and
			 * off-site), then the question is what to do with the outbound mail
			 * addresses: 
			 *
			 * a) drop addresses that can't be transformed; not ideal since some
			 * will receive the message in a spam folder and others won't recieve
			 * it at all.
			 *
			 * b) forward to the system collection box; this is more in line with
			 * previous behaviour of the milter and allows the postmaster to 
			 * retrain or whitelist within SpamAssassin.
			 *
			 * c) reject the whole message outright; this is a nice choice, since
			 * the sender gets immediate feedback that there is a problem, but 
			 * not so easy to retrain SpamAssassin.
			 *
			 * d) don't scan outbound mail; this involves determining whether the
			 * sender is truly a local user (login account, aliases database, 
			 * virtual user box), which increases the code complexity. See -f option.
			 */
			int sendToSystemSpamBox = redirectSpam != REDIRECT_TO_FORMAT || VectorLength(data->rcpts) != data->localRcptCount;
		
			if (redirectSpam) {
				/* Remove previous list of recipients from Sendmail's list
				 * and record them in the redirected message in case it was
				 * not spam and needs to be resent later. Sometimes the To:
				 * or Cc: headers are useless ie. undisclosed recipients.
				 */
				for (i = 0; i < VectorLength(data->rcpts); i++) {
					if ((rcpt = VectorGet(data->rcpts, i)) == NULL)
						continue;

					if (smfi_addheader(ctx, X_ORIGINAL_RECIPIENT, rcpt->address.string) == MI_SUCCESS)
						(void) smfi_delrcpt(ctx, rcpt->address.string);
				}

				/* If -F was given, then modify the recipient list so that
				 * each local recipient is modified according to a sepcified
				 * format.
				 */
				if (!sendToSystemSpamBox) {
					for (i = 0; i < VectorLength(data->rcpts); i++) {
						if ((rcpt = VectorGet(data->rcpts, i)) == NULL)
							continue;
							
						/* We cannot modify non-local recipients, because
						 * plus details are a sendmail thang and the remote
						 * end might handle spam differently.
						 */
						if (!rcpt->isLocal)
							continue;
						
						(void) formatPath(data->line, sizeof(data->line), rcptReformat, rcpt);
					
						if (smfLogDetail & SMF_LOG_TRACE)
							syslog(LOG_DEBUG, TAG_FORMAT "modified recipient=%s", TAG_ARGS, data->line);
						
						if (smfi_addrcpt(ctx, data->line) == MI_FAILURE)
							syslog(LOG_ERR, TAG_FORMAT "redirect-spam-to %s failed ", TAG_ARGS, data->line);
					}
				}
			}

			if (sendToSystemSpamBox && smfi_addrcpt(ctx, copySpamTo) == MI_FAILURE) {
				syslog(LOG_ERR, TAG_FORMAT "copy-spam-to <%s> failed ", TAG_ARGS, copySpamTo);
				goto error1;
			}
		}
	} else {
		/* Send a copy of clean mail to an archive address. */
		if (copyHamTo != (char *) 0 && smfi_addrcpt(ctx, copyHamTo) == MI_FAILURE)
			syslog(LOG_ERR, TAG_FORMAT "copy-ham-to <%s> failed ", TAG_ARGS, copyHamTo);
	}

error1:
	/* REUSE this buffer to build RCPT list string. */
	BufSetLength(data->headers, 0);

	/* Save copy of recipient list into a string for log information. */
	for (i = 0; i < VectorLength(data->rcpts); i++) {
		if ((rcpt = VectorGet(data->rcpts, i)) != NULL) {
			BufAddString(data->headers, rcpt->address.string);
			BufAddString(data->headers, ">,<");
		}
	}
	
	/* Remove trailing ",<". */
	BufSetLength(data->headers, BufLength(data->headers)-2);

	syslog(
		LOG_INFO, TAG_FORMAT "spam=%s score=%.2f required=%.2f client_addr=%s client_name=%s subject='%s' mail=<%s> rcpts=<%s",
		TAG_ARGS, isSpam ? "YES" : "NO", score, threshold, 
		data->client_addr, data->client_name, data->subject,
		data->work.mail->address.string, BufBytes(data->headers)
	);		

	if (isSpam) {
		(void) smfReply(&data->work, 550, "5.7.1", "message was identified as junk mail, score %.2f/%.2f", score, threshold);

		if (threshold + rejectExcessScore <= score)
			return (checkPrecedenceList && data->precedence <= PRECEDENCE_LIST) ? SMFIS_DISCARD : SMFIS_REJECT;
		if (threshold + discardExcessScore <= score)
			return SMFIS_DISCARD;
#ifdef SMFIS_REJECT_AND_DELIVER
		if (threshold + rejectAndDeliverExcessScore <= score)
			return SMFIS_REJECT_AND_DELIVER;
#endif
	}
error0:
	return SMFIS_CONTINUE;
}

/*
 * Close and release per-connection resources.
 */
static sfsistat
filterClose(SMFICTX *ctx)
{
	workspace data;
	unsigned short cid;

	if ((data = (workspace) smfi_getpriv(ctx)) == NULL)
		return smfNullWorkspaceError("filterClose");

	cid = smfCloseEpilog(&data->work);

	if (data->localRcpt != spamdUser)
		free(data->localRcpt);

	VectorDestroy(data->rcpts);
	BufDestroy(data->headers);
	SocketClose(data->server);
	free(data);

	if (smfLogDetail & SMF_LOG_TRACE)
		syslog(LOG_DEBUG, TAG_FORMAT "filterClose(%lx)", cid, smfNoQueue, (long) ctx);

	return SMFIS_CONTINUE;
}

/***********************************************************************
 ***  Milter Definition Block
 ***********************************************************************/

static struct smfiDesc smfilter = {
	MILTER_NAME,		/* filter name */
	SMFI_VERSION,		/* version code -- do not change */
	SMFIF_ADDHDRS|SMFIF_CHGHDRS|SMFIF_ADDRCPT|SMFIF_DELRCPT,	/* flags */
	filterOpen,		/* connection info filter */
	filterHelo,		/* SMTP HELO command filter */
	filterMail,		/* envelope sender filter */
	filterRcpt,		/* envelope recipient filter */
	filterHeader,		/* header filter */
	filterEndHeaders,	/* end of header */
	filterBody,		/* body block filter */
	filterEndMessage,	/* end of message */
	NULL,			/* message aborted */
	filterClose,		/* connection cleanup */
};

/***********************************************************************
 *** Startup
 ***********************************************************************/

static void
atExitCleanUp()
{
	smdbClose(smdbAccess);
	smfAtExitCleanUp();
}

int
main(int argc, char **argv)
{
	int ch, rc;
	struct stat sb;

	openlog(MILTER_NAME, LOG_PID, LOG_MAIL);
	(void) smfi_settimeout(SMF_SOCKET_TIMEOUT);

	alwaysAddReport = 0;
	copyAllTo = (char *) 0;
	copyHamTo = (char *) 0;
	copySpamTo = (char *) 0;
	socketTimeout = DEFAULT_SOCKET_TIMEOUT;
	subjectTag = DEFAULT_SPAM_SUBJECT_TAG;
	appendScoreToSubjectTag = 0;
	spamdUser = (char *) 0;
	forceSpamdUser = 0;
	redirectSpam = COPY_TO;
	untrustedRcptLimit = 0;
	spamcCommand = DEFAULT_SPAMC_COMMAND;
	rejectExcessScore = ~(unsigned long)0 >> 1;
	discardExcessScore = ~(unsigned long)0 >> 1;
#ifdef SMFIS_REJECT_AND_DELIVER
	rejectAndDeliverExcessScore = ~(unsigned long)0 >> 1;
#endif

	while ((ch = getopt(argc, argv, getoptString)) != -1) {
		switch (ch) {
		case 'A':
			alwaysAddReport = 1;
			spamcCommand = "REPORT";
			break;
		case 'C':
			alwaysAddReport = 0;
			spamcCommand = "CHECK";
			break;
		case 'a':
			copyAllTo = optarg;
			break;
		case 'b':
			copySpamTo = optarg;
			redirectSpam = COPY_TO;
			break;
		case 'B':
			copySpamTo = optarg;
			redirectSpam = REDIRECT_TO;
			break;
		case 'F':
			rcptReformat = optarg;
			redirectSpam = REDIRECT_TO_FORMAT;
			break;
		case 'c':
			copyHamTo = optarg;
			break;
		case 'd':
			discardExcessScore = strtol(optarg, (char **) 0, 10);
			if (discardExcessScore < 0) {
				(void) fprintf(stderr, "-d value must be positive");
				(void) fprintf(stderr, usageMessage);
			}
			break;
		case 'f':
			sendmailCf = optarg;
			break;
		case 'K':
			maxBytes = strtol(optarg, (char **) 0, 10);
			break;			
		case 'l':
			untrustedRcptLimit = strtol(optarg, (char **) 0, 10);
			if (untrustedRcptLimit < 0) {
				(void) fprintf(stderr, "-l value must be positive");
				(void) fprintf(stderr, usageMessage);
			}
			break;
		case 'T':
			socketTimeout = strtol(optarg, (char **) 0, 10);
			break;
		case 't':
			(void) smfi_settimeout((int) strtol(optarg, NULL, 10));
			break;
		case 'p':
			checkPrecedenceList = 1;
			break;
		case 'r':
			rejectExcessScore = strtol(optarg, (char **) 0, 10);
			if (rejectExcessScore < 0) {
				(void) fprintf(stderr, "-r value must be positive");
				(void) fprintf(stderr, usageMessage);
			}
			break;
#ifdef SMFIS_REJECT_AND_DELIVER
		case 'R':
			rejectAndDeliverExcessScore = strtol(optarg, (char **) 0, 10);
			if (rejectAndDeliverExcessScore < 0) {
				(void) fprintf(stderr, "-R value must be positive");
				(void) fprintf(stderr, usageMessage);
			}
			break;
#endif
		case 's':
			subjectTag = optarg;
			break;
		case 'S':
			appendScoreToSubjectTag = 1;
			break;
		case 'u':
			spamdUser = optarg;
			forceSpamdUser = 0;
			break;
		case 'U':
			spamdUser = optarg;
			forceSpamdUser = 1;
			break;
		case 'v':
			(void) smfSetLogDetail(optarg);
			break;
		case 'H':
			spamdHost = optarg;
			break;
		case 'P':
			spamdPort = (int) strtol(optarg, (char **) 0, 10);
			break;
		default:
			syslog(LOG_ERR, "unknown command line option -%c", optopt);
			/*@fallthrough@*/
		case 'h':
		case '?':
			(void) fprintf(stderr, usageMessage);
			return 2;
		}
	}

	if (optind + 1 != argc) {
		syslog(LOG_ERR, "missing command line argument");
		(void) fprintf(stderr, usageMessage);
		return 2;
	}

	milter.socket = argv[optind];	
	socketTimeout *= 1000;
	maxBytes *= 1024;
	
	if (atexit(atExitCleanUp)) {
		syslog(LOG_ERR, "atexit() failed\n");
		return 1;
	}

	/* Specifying the -s option enable access and aliases database support. 
	 * NOTE that if we are running as root at this point, we'll successfully
	 * open the database with those privelages, then below the process
	 * owneship will change. Any changes to the access.db that triggers a
	 * reload will be done using lesser privelages.
	 */
	if (sendmailCf != NULL && *sendmailCf != '\0') {
		if (smfLogDetail & SMF_LOG_DATABASE)
			smdbSetDebugMask(SMDB_DEBUG_ALL);

		(void) readSendmailCf(sendmailCf, SMDB_OPEN_ACCESS);
		
		if (smdbAccess == NULL || smdbAccess->db == (DB *) 0)
                        syslog(LOG_WARNING, "access database support not available");
	}

	if (smfLogDetail & (SMF_LOG_SOCKET_FD|SMF_LOG_SOCKET_ALL))
		SocketSetDebugMask(
			(smfLogDetail & SMF_LOG_SOCKET_ALL) 
				? SOCKET_DEBUG_ALL
				: SOCKET_DEBUG_FD
		);

	if (SocketInit()) {
		syslog(LOG_ERR, "SocketInit() error\n");
		return 1;
	}

	return smfMainStart(&milter, smfilter);
}
